#!/usr/bin/env perl

use strict;
use Data::Dumper;
use List::Util qw(max);
use XML::RPC;

my $VERSION = "2019.07.31";
my $DEBUG = $ENV{HL_DEBUG} ? 1 : 0;

my $vserver_data;
my $rootdir = "/var/lib/hurtlocker";

my $ERROR = -1;
my $SIZE_OK = 0;
my $SIZE_UP = 1;
my $KILLED = 2;
my $SIZE_DOWN = 3;

my $GENEROUS_RAM_DEFAULT = 1000 * 1024 * 1024;

# force here means we should size up even if we wouldn't need to based on
# current usage
my ($vserver, $action_type, $force) = @ARGV;
usage() && exit($ERROR) if !$vserver || !$action_type || (defined $force && $force ne '--force');

my $message = '';

sub print_version {
	print $VERSION . "\n";
	return 1;
}

sub log_xmlrpc {
	my $action = shift;
        my $xmlrpc = XML::RPC->new("http://xmlrpc.dreamhost.com/index.fcgi");
	my $ret = 0;

	$ret = $xmlrpc->call("Hurtlock.hurt", $vserver, $action, $message);
	write_to_logfile(Dumper($ret)) if $DEBUG;
}

sub log_flatfile {
	my ($action) = shift;

	write_to_logfile($message);
}

sub write_to_logfile {
	my ($text) = shift;

	open(my $FILE, ">>", "/var/log/hurtlocker.log");
	print $FILE localtime() . ": $vserver: $text\n";
	close($FILE);
	return;
}

sub log_all {
	my ($action) = shift;

	log_flatfile($action);
	log_xmlrpc($action);
}

sub usage {
	print <<'USAGE'
USAGE: cgroup_vserver_handler.pl <vserver name> <command> [<options>]

commands: vkill, resize, dump, version

options: when using the resize command, --force may be specified to force
resizing up even when we would deem it unnecessary.  This will not force
resizing over normal resize limits.
USAGE
}

sub size_up {
	my ($base_lim, $new_lim) = @_;

	write_to_logfile("Sizing up... New limit will be: $new_lim") if $DEBUG;

	system("echo $new_lim > /sys/fs/cgroup/memory/$vserver/memory.memsw.limit_in_bytes");
	system("echo $new_lim > /sys/fs/cgroup/memory/$vserver/memory.soft_limit_in_bytes");
	system("echo $new_lim > /sys/fs/cgroup/memory/$vserver/memory.limit_in_bytes");

	open(my $RESIZE_FILE, '>', "$rootdir/$vserver.resize");
	print $RESIZE_FILE $base_lim;
	$message .= " resized up to $new_lim";
	log_all("up");
	return 1;
}

sub size_down {
	my ($new_lim) = @_;

	write_to_logfile("Sizing down... New limit will be: $new_lim") if $DEBUG;

	system("echo $new_lim > /sys/fs/cgroup/memory/$vserver/memory.limit_in_bytes");
	system("echo $new_lim > /sys/fs/cgroup/memory/$vserver/memory.soft_limit_in_bytes");
	system("echo $new_lim > /sys/fs/cgroup/memory/$vserver/memory.memsw.limit_in_bytes");

	$message .= " resized down to $new_lim";
	log_all("down");
	return 1;
}

sub dump_process_list {

	open(FILEOUT, "> $rootdir/$vserver.dump") || return -1;
	my @procdata = `vps faxu |grep $vserver |grep -v "grep $vserver" | grep -v "$0"`;
	foreach my $line (@procdata){
		print FILEOUT $line;
	}
	close(FILEOUT);
}

sub clean_artifacts {
	my @file_ends = @_;
	@file_ends = qw(resize generous_ram) if !@file_ends;
	for my $file_end (@file_ends) {
		my $artifact = "$rootdir/$vserver.$file_end";
		unlink $artifact if -e $artifact;
	}
	return;
}

sub read_first_line {
	my ($file_name) = @_;
	open(my $fh, '<', $file_name) or exit $ERROR;
	my $first_line = <$fh>;
	return $first_line;
}

sub resize {
	#don't resize if vserver is not currently running
	my $is_up = read_first_line("/sys/fs/cgroup/cpuset/$vserver/cgroup.procs");

	if($is_up !~ /\d+/){
		write_to_logfile("$vserver is not currently running") if $DEBUG;
		clean_artifacts();

		return $ERROR;
	}

	#get current limit
	my $cur_lim = int(read_first_line("/sys/fs/cgroup/memory/$vserver/memory.limit_in_bytes"));
	my $base_lim = $cur_lim;

	# if previously resized, base limit == 75% of the current limit (the size of resize_artifact)
	my $resize_artifact = "$rootdir/$vserver.resize";
	if (-e $resize_artifact) {
		$base_lim = int(read_first_line($resize_artifact)) || ($cur_lim * 0.8);
	}

	my $new_lim = int($base_lim * 1.25);

	# i've only ever seen generous ram applied to NDN accounts
	my $generous_artifact = "$rootdir/$vserver.generous_ram";
	if ( -e $generous_artifact ) {
		my $generous_ram_amount =
			int(read_first_line($generous_artifact)) || $GENEROUS_RAM_DEFAULT;
		$new_lim = max($new_lim, $base_lim + $generous_ram_amount);
	}
	# NB: this means that new_lim could be less than, equal to, or greater than
	# cur_lim.  The important part is that as of the current state of the
	# world, this is the limit up for consideration, and that may be lower than
	# what the current limit is.

	write_to_logfile(
		"base limit: $base_lim; cur lim: $cur_lim; resized would be: $new_lim"
	) if $DEBUG;

	#get system usage... if it is at 80% memory or more, kill instead of resize
	#TODO: future version.

	#get current vserver usage
	my $cur_use = int(read_first_line("/sys/fs/cgroup/memory/$vserver/memory.usage_in_bytes"));

	write_to_logfile("cur usage: $cur_use") if $DEBUG;
	$message .= "Limit $cur_lim with usage of $cur_use";

	# vkill if current usage is greater than 125% of the value of memory.limit_in_bytes
	if ($cur_use > $new_lim) {
		write_to_logfile("Going to kill it anyway, it's thrashing too hard, sorry") if $DEBUG;
		vkill_it();
		clean_artifacts();
		return $KILLED;
	} elsif (-e $resize_artifact) {
		#already been resized.
		write_to_logfile("Found already resized from $base_lim to $cur_lim") if $DEBUG;
		if ($base_lim > $cur_use && !$force) {
			size_down($base_lim);
			clean_artifacts('resize');
			return $SIZE_DOWN;
		} else {
			# Not big enough to kill, but not small enough to go back, reset
			# lim to whatever is correct right now, which could mean going up
			# or down, or just staying.  This allows hurtlocker to be
			# idempotent.

			# NOTE: the event_listener wont trigger a call to this resize() module
			# if the resize_artifact exists
			if ($cur_lim == $new_lim) {
				write_to_logfile("Staying at current limit") if $DEBUG;
				return $SIZE_OK;
			} elsif ($new_lim > $cur_lim) {
				size_up($base_lim, $new_lim);
				return $SIZE_UP;
			} else {
				size_down($new_lim);
				return $SIZE_DOWN;
			}
		}
	# Resize at 90%+ RAM usage
	} elsif ($cur_use >= ($base_lim * 0.90) && !$force) {
		# vserver should resize  even if 100% memory limit is never hit. SEC-1707
		write_to_logfile("Resizing since memory usage exceeded 90%") if $DEBUG;
		size_up($base_lim, $new_lim);
		return $SIZE_UP;
	} elsif ($base_lim > $cur_use && !$force) {
		# We called resize, but we didn't need to resize
		write_to_logfile("Staying at current limit") if $DEBUG;
		return $SIZE_OK;
	} else {
		size_up($base_lim, $new_lim);
		return $SIZE_UP;
	}
}

sub vkill_it {

	my $context_id = `cat /etc/vservers/$vserver/context`;
	chomp($context_id);

	if (! -f "/proc/virtual/$context_id/status") {
		return $ERROR;
	} elsif ($DEBUG) {
		write_to_logfile("in vkill_it()");
		my ($status) = `cat /proc/virtual/$context_id/status`;
		write_to_logfile("status: $status");
	}

	#will rewrite this in a different way next week
	#using vps instead of walking /proc will cause problems
	#if a vserver is hung.
	#dump_process_list();

	my $status = 0;
	$status = system("vkill -s 9 -c $context_id");
	$message .= " vkilled";
	if ($status != -1 && $status < 256){
		# We send a restart vs start, since the vkill doesn't always stop the instance
		# and a restart starts a stopped instance.
		system("echo /usr/sbin/vserver $vserver restart | at now+1min");
		$message .= " and scheduled to restart";
	}
	log_all('killed');

	open(FILE, "/etc/vservers/$vserver/cgroup/memory.limit_in_bytes") or return $ERROR;
	my $mem_lim = <FILE>;
	close(FILE);

	#explicit checks for these files, just in case vkill did the right thing and
	#removed the vserver directories, in which case of course we don't want to write
	#to files inside of them.
	if ( -e "/sys/fs/cgroup/memory/$vserver" ) {
		system("cp /etc/vservers/$vserver/cgroup/memory.memsw.limit_in_bytes /sys/fs/cgroup/memory/$vserver");
		system("cp /etc/vservers/$vserver/cgroup/memory.limit_in_bytes /sys/fs/cgroup/memory/$vserver");
		system("cp /etc/vservers/$vserver/cgroup/memory.memsw.limit_in_bytes /sys/fs/cgroup/memory/$vserver");
	}
	system("echo 1 > /etc/vservers/$vserver/cgroup/memory.oom_control");
	return $KILLED;

}

my $retval = 0;
write_to_logfile("Going to $action_type $vserver") if $DEBUG;
if($action_type =~ /vkill/){
	$retval = vkill_it();
} elsif ($action_type =~ /resize/){
	$retval = resize();
} elsif ($action_type =~ /dump/){
	$retval = dump_process_list();
} elsif ($action_type =~ /version/){
	$retval = print_version();
}
exit $retval;

