#!/usr/bin/env perl

use strict;
use warnings;
use feature 'say';
use File::ChangeNotify;
use File::Copy 'cp';
use File::Path 'make_path';
use File::Pid;
use File::Slurp;
use Getopt::Long;
use Math::BigFloat ':constant';
use POSIX 'strftime';

my ($threshold, $kill_limit, $sleep_int, $pid_file, $debug, $nokill);

GetOptions (
	"h|help"         => sub { usage(); },
	"t|threshold=n"  => \$threshold,
	"k|kill_limit=n" => \$kill_limit,
	"s|sleep_int=n"  => \$sleep_int,
	"p|pid=s"        => \$pid_file,
	"d|debug"        => \$debug,
	"n|nokill"       => \$nokill,
);

$debug = 1 if $nokill;

$threshold  ||= ".98";
$kill_limit ||= 6;
$sleep_int  ||= 10;
$pid_file   ||= "/var/run/resize_listener.pid";

my $vserver_path = "/var/lib/vservers/";
my $watch_dir  = "/var/lib/hurtlocker/";
my $vserver_memory_dir = "/sys/fs/cgroup/memory/";
my $vserver_memory_limit_file = "/memory.memsw.limit_in_bytes";
my $vserver_memory_usage_file = "/memory.memsw.usage_in_bytes";
my $mem_hog_bin = "/dh/bin/hurtlocker/mem-hog";
my $vservers_path = "/var/lib/vservers/";
my $guest_mem_hog_loc = "/root/mem-hog";

my $pidfile = File::Pid->new({
  file => $pid_file,
});

if ( my $num = $pidfile->running ) {
	die "script already running pid: $num\n";
}

$pidfile->write;

my %Vservers;

my $watcher = File::ChangeNotify->instantiate_watcher(
		directories => $watch_dir,
		filter      => qr/\.(resize)$/,
	);

say "Initalizing...";
write_to_logfile("Initalizing...");
while (1) {
	if ( my @events = $watcher->new_events ) {

		say "\n== collecting events" if $debug;
		for my $event (@events) {
			my $resize = (split '/', $event->path)[-1];
			my $vps = substr($resize, 0, index($resize, "."));

			if ($event->type eq 'delete') {
				delete($Vservers{$vps});
				say "$vps: resize file was removed from hash" if $debug;
				write_to_logfile("$vps: resize file was removed, not watching it anymore");
				next;
			} else {
				say "$vps: watching | " . $event->path if $debug;
				write_to_logfile("$vps: resize file found, watching it");
			}

			my $resized = read_file($event->path);

			if (! -e $vserver_memory_dir . $vps . $vserver_memory_limit_file) {
				say "$vps: no memory limit file exists!" if $debug;
				next;
			}
			if (! -e $vserver_memory_dir . $vps . $vserver_memory_usage_file) {
				say "$vps: no memory usage file exists!" if $debug;
				next;
			}
			$Vservers{$vps}{resized} = 1;
		}
	} else {
		say "waiting for event, not watching any vservers at the moment... sleeping" if !%Vservers && $debug;
	}

	say "\n= processing vservers" if %Vservers && $debug;
	for my $vserver (sort keys %Vservers) {
		my $vps_limit = read_file($vserver_memory_dir . $vserver . $vserver_memory_limit_file);
		my $vps_usage = read_file($vserver_memory_dir . $vserver . $vserver_memory_usage_file);
		$vps_limit = int($vps_limit);
		$vps_usage = int($vps_usage);
		my $hr_vps_limit = FileSize($vps_limit) if ($vps_limit >= 0);
		my $hr_vps_usage = FileSize($vps_usage) if ($vps_usage >= 0);
		my $thresh = int($vps_limit * $threshold);
		my $hr_thresh = FileSize($thresh) if ($thresh >= 0);
		my $tick      = $Vservers{$vserver}{tick};
		$tick ||= 1;

		say "$vserver: $hr_vps_usage/$hr_vps_limit | threshold: $hr_thresh" if $debug;
		if ($vps_usage >= $thresh) {
			say "\tabove the threshhold in memory: $hr_vps_usage >= $hr_thresh" if $debug;
			print "\t$tick/$kill_limit" if $debug;
			if ($tick == $kill_limit) {
				print " times up!" if $debug;
				print "\t needs to be killed\n" if $debug;
				write_to_logfile("$vserver: above memory threshold $hr_thresh for $kill_limit ticks, needs to be oomed and or killed");
				cp($mem_hog_bin, $vservers_path.$vserver.$guest_mem_hog_loc);
				unless ($nokill) {
					if ( -e $vservers_path.$vserver.$guest_mem_hog_loc) {
						if (system("/usr/sbin/vserver $vserver exec /root/mem-hog")) {
							write_to_logfile("$vserver: executed mem_hog");
						}
					} else {
						write_to_logfile("$vserver: mem_hog dosnt exist on guest, cant trigger OOM event");
					}
				}
			} elsif ($tick > $kill_limit) {
				write_to_logfile("$vserver: restarting guest, mem-hog did not trigger");
				system("/usr/sbin/vserver $vserver restart");
			} else {
				print " tick-tock\n" if $debug;
				$Vservers{$vserver}{tick}++;
				write_to_logfile("$vserver: above memory threshold ($hr_vps_usage >= $hr_thresh) for $tick ticks");
			}
		} else {
			$Vservers{$vserver}{tick} = 0;
		}
	}
	sleep $sleep_int;
}

## Taken from Ndn/Common/Formatting.pm

sub Decimal {
	my $dec = $_[1];
	$dec = 2 unless defined $_[1];
	sprintf("%.".$dec."f",$_[0]);
}

=item FileSize($value)

Returns 1073741824 as 1GB and other filesizes (in Bytes) as a nice readable amount.

=cut

sub FileSize {
	my $value = $_[0];
	#10737418240
	return $value. ' B' if $value < 1024;
	$value /= 1024; #10485760
	return Decimal($value,$_[1]). ' KB' if $value < 1024;
	$value /= 1024; #10240
	return Decimal($value,$_[1]). ' MB' if $value < 1024;
	$value /= 1024;
	return Decimal($value,$_[1]). ' GB'; # guests should not a TB of ram
}

sub write_to_logfile {
	my ($text) = shift;
	return unless $debug;
	unless(-e "/var/log/hurtwatcher/" or mkdir("/var/log/hurtwatcher/", 0755)) {
		die "Unable to create log directory /var/log/hurtwatcher/\n";
	}
	write_file("/var/log/hurtwatcher/hurtwatcher.log", {append => 1}, $text . "\n");
	return;
}

sub usage {

	my $msg = shift;
	print "[ERROR] $msg\n" if $msg;

	print << "EOF";
[About]
  Listens for resize file created by hurtlocker (/var/lib/hurtlocker/). If
  machine is found to be resized it will start watching the memory usage on the
  guest. If the memory usage is found to be at the memory threshold for
  (default) 6 'ticks' it will trigger an exec mem-hog on behalf of the vserver,
  forcing it to trigger an oom_event hurtlocker catches the oom_event and with
  the existence of the resize file, vkills it.
  Threshold is 98% of the limit in /sys/fs/cgroup/memory/\$ps/memory.memsw.limit_in_bytes.

  note:
    a 'tick' is 10 seconds.

[Usage]
  $0 [--help|--threshold|--kill_limit|--sleep_int|--pid|--debug]

[Flags]

  -h, --help       - this help
  -t, --threshold  - memory % threshold. default: .98 (98%)
  -k, --kill_limit - trigger oom killer script after n amount of ticks. default: 6
  -s, --sleep_int  - sleep time before next tick default: 10 (seconds)
  -p, --pid        - location of pid file default: /var/run/resize_listener.pid
  -d, --debug      - additional print statements for troubleshooting
  -n, --nokill     - dont kill anything, just output. invokes --debug

EOF
	$pidfile->remove;
	$msg ? exit 1 : exit 0;

}

say "Goodbye";
$pidfile->remove;
write_to_logfile("Exited...");

return;
